#include <stdio.h>
#include "parser.h"
#include "opt.h"
#include "findrun.h"

void makecutrow(signed char *temp, signed char *block, int length);

void makecuts(int start, int end)
{
    int b;
    for (b = 0; b < BLOCKCOUNT; b ++)
        makecutrow(&o_temp[b][start], &o_block[b][start], end - start);
}

void makecutrow(signed char *temp, signed char *block, int length)
{
    int i, j, k;

    if (length == 0) return;
/*
    fprintf(stderr, "makecutrow\n");
    fprintf(stderr, "in:  ");
    for (i = 0; i < length; i ++)
        fprintf(stderr, " %2d", block[i]);
    fprintf(stderr, "\n");
*/
    for (i = 0; i < length; i ++) {
        if (block[i] != 1) temp[i] = 0;
        else {
            for (; i < length && block[i] == 1; i ++) temp[i] = 1;
            if (i >= length) break;
            if (block[i] != -1) {
                temp[i] = 0;
                continue;
            }
            for (j = i; j < length && block[j] == -1; j ++);
            if (j < length && block[j] == 1) {
                for (k = i; k < j; k ++) temp[k] = 1;
            }
            i = j - 1;
        }
    }
/*
    fprintf(stderr, "out: ");
    for (i = 0; i < length; i ++)
        fprintf(stderr, " %2d", temp[i]);
    fprintf(stderr, "\n");
*/
}

void updatecuts(int cutpos, int start, int finish)
{
    int b, i;
    /*fprintf(stderr, "updatecuts(%d, %d, %d)\n", cutpos, start, finish);*/
    if (cutpos == start || cutpos == finish) return;
    for (b = 0; b < BLOCKCOUNT; b ++) {
        for (i = cutpos; i < finish && o_temp[b][i] && o_block[b][i] != 1; i ++)
            o_temp[b][i] = 0;
        for (i = cutpos-1; i >= start && o_temp[b][i] && o_block[b][i] != 1; i --)
            o_temp[b][i] = 0;
    }
}

int scorecut(int pos, int start, int finish)
{
    int score, i;
    score = 0;
    if (pos == start || pos == finish)
        return 0;
    for (i = 0; i < BLOCKCOUNT; i ++)
        if (o_temp[i][pos] == 1 && o_temp[i][pos-1] == 1)
            score ++;
    return score;
}

int squashrun(int start, int end, int b, int *b_start, int *b_end,
              int rangestart, int rangeend)
{
    int onestart, oneend, s, e, bs, be, bs_s, be_s, score;

    /*fprintf(stderr, "squashrun(%d, %d, %d) .. \n", start, end, b);

    {
        int i;
        for (i = start; i < end; i ++)
            fprintf(stderr, " %d", o_block[b][i]);
        fprintf(stderr, "\n");
    }*/

    for (onestart = start; onestart < end &&
            o_block[b][onestart] == -1; onestart ++);
    for (oneend = end; oneend > start &&
            o_block[b][oneend-1] == -1; oneend --);

    /*fprintf(stderr, "%d %d\n", onestart, oneend);*/

    if (oneend <= onestart) {
        /*fprintf(stderr, "none\n");*/
        return -1;
    }

    bs_s = bs = -1;
    for (s = start; s <= onestart; s ++) {
        score = scorecut(s, rangestart, rangeend);
        /*fprintf(stderr, "s=%d score=%d\n", s, score);*/
        if (bs_s == -1 || score < bs_s) {
            bs_s = score;
            bs = s;
        }
    }

    be_s = be = -1;
    for (e = end; e >= oneend; e --) {
        score = scorecut(e, rangestart, rangeend);
        /*fprintf(stderr, "e=%d score=%d\n", e, score);*/
        if (be_s == -1 || score < be_s) {
            be_s = score;
            be = e;
        }
    }

    /*fprintf(stderr, "start=%d end=%d score=%d\n", bs, be, bs_s + be_s);*/

    *b_start = bs;
    *b_end = be;
    return bs_s + be_s;
}

TAG findrun(int start, int finish, int *runstart, int *runend)
{
    int b, i, j, score, b_start, b_end;
    int m_b, m_start, m_end, m_score;
    int mid = (start+finish)/2;

    m_score = m_start = m_end = m_b = -1;

    /*fprintf(stderr, "findrun(%d, %d) .. \n", start, finish);*/

    for (b = 0; b < BLOCKCOUNT; b ++) {
        for (i = start; i < finish; i ++) {
            if (o_block[b][i]) {
                for (j = i; j < finish && o_block[b][j]; j ++);
                score = squashrun(i, j, b, &b_start, &b_end, start, finish);
                i = j;
                if (score == -1) continue;
                if (m_score == -1 || score < m_score || (score == m_score && j < mid)) {
                    m_score = score;
                    m_start = b_start;
                    m_end = b_end;
                    m_b = b;
                }
            }
        }
    }

    if (m_score == -1) {
        /*fprintf(stderr, "none\n");*/
        return T_NONE;
    }

    /*fprintf(stderr, "start=%d end=%d b=%d score=%d\n", m_start, m_end, m_b, m_score);*/

    for (i = m_start; i < m_end; i ++) {
        o_block[m_b][i] = 0;
        o_temp[m_b][i] = 0;
    }

    *runstart = m_start;
    *runend = m_end;
    return o_taglist[m_b];
}
